from nodoc.post_processing import ocrData
from nodoc.document import Message
from nodoc.document import Checker

"""
A special program which processing the last flow.
"""

def parse(data: ocrData) -> Message:
    """
    Parse the message from the last flow.
    - data: ocrData,  used to parse to normal message.
    """
    final_info: list[tuple[str, float, float]] = []
    # 解码内容，左右总字符距，上下字符距
    message: Message = {
        "metadata": None,
        "pipeline": []
    }
    fontSizes = set([
        unit['text_region'][2][1] - unit['text_region'][0][1] 
        for element in data 
        for unit in element['res']
        if element['type'] == 'title'
    ])
    fontSizes = list(fontSizes)
    fontSizes.sort()
    for element in data:
        if element['type'] == 'figure': 
            continue
        for unit in element['res']:
            if element['type'] == 'text':

                fontHeight: str = unit['text_region'][2][1] - unit['text_region'][0][1] # 字符高度
                textWidth: str = unit['text_region'][2][0] - unit['text_region'][0][0]  # 文本宽度
                indent: float = unit['text_region'][0][0] - element['bbox'][0]          # 字符缩进（像素）
                tailIndent: float = unit['text_region'][0][0] - element['bbox'][0]      # 文本尾部内边距（像素）
                content: str = u"" + unit['text']                                       # 内容
                charCount = len(content)
                message['pipeline'].append({
                    'type': element['type'],
                    'property': {
                        'indent':  indent,
                        'inherited': {
                            'font_size': fontHeight,
                            'font_distance': textWidth / charCount - fontHeight,
                            'position': [
                                unit['text_region'][0][0], unit['text_region'][2][0],
                            ],
                            'width': textWidth,
                            'height': fontHeight
                        }
                    },
                    'content': content
                })

            elif element['type'] == 'title':

                fontHeight: str = unit['text_region'][2][1] - unit['text_region'][0][1] # 字符高度
                textWidth: str = unit['text_region'][2][0] - unit['text_region'][0][0]  # 文本宽度
                indent: float = unit['text_region'][0][0] - element['bbox'][0]          # 字符缩进（像素）
                tailIndent: float = unit['text_region'][0][0] - element['bbox'][0]      # 文本尾部内边距（像素）
                content: str = u"" + unit['text']                                       # 内容
                charCount = len(content)
                message['pipeline'].append({
                    'type': element['type'],
                    'property': {
                        'inherited': {
                            'font_size': fontHeight,
                            'font_distance': textWidth / charCount - fontHeight,
                            'position': [
                                unit['text_region'][0][0], unit['text_region'][2][0],
                            ],
                            'width': textWidth,
                            'height': fontHeight
                        },
                        'level': fontSizes.index(fontHeight) + 1
                    },
                    'content': content,
                })
    return message



import os
import re
from .base import Document



class PDF(Document):

    def __init__(self, data: str = "") -> None:
        message = self.normalize(data)
        super().__init__(message, data)
        
    def normalize(self):
        ...

    def export(self, name: str, directory: str = './'):
        return super().export(name, directory)


import re
from typing import Literal, TypeAlias, Self, Union

import torch
from .database import dataBase
from nodoc.const import Embedding
from sentence_transformers import SentenceTransformer
import numpy as np
import os
from nodoc import debugger
import faiss
import ctypes
from nodoc import const


def __import():
    global docNode, docTree
    from nodoc import docNode
    from nodoc import docTree


_docNodes: TypeAlias = list['docNode']
_Embeddings: TypeAlias = np.ndarray[Embedding]
_Forest: TypeAlias = list['docTree']


class c_methods:

    @staticmethod
    def search(query_matrix: np.ctypeslib._ndptr, query_text: np.ctypeslib._ndptr, rows: int, columns: int) -> int:
        """
        搜索查询矩阵与query_text最相似的向量的索引。
        - query_matrix: _ndptr, 传入一个查询矩阵的指针。
        - query_text: _ndptr, 传入待查询文本嵌入的指针。
        - rows: int, 行数。
        - columns: 列数。
        返回一个整型索引值。
        """


class vectorDB(dataBase):

    def __init__(self, forest: _Forest = [], mode: Literal['exact', 'low'] = 'exact', cache_folder='./', *, model: SentenceTransformer | None = None) -> None:
        """
        实例化一个向量数据库对象。
        - forest: _Forest, 树列表，作为数据库的查询树。
        - mode: Literal['exact', 'low'], `exact`为准确模式，`low`为低耗模式，决定数据库使用的Embedding模型。
        - cache_folder: str = './', 数据库模型的缓存路径。
        """
        super().__init__()
        if model is None:
            match mode:
                case 'low':
                    __model = 'BAAI/bge-small-zh'
                case _:
                    __model = 'BAAI/bge-large-zh-v1.5'
            device = 'cuda' if torch.cuda.is_available() else None
            self.model = SentenceTransformer(
                __model, cache_folder=cache_folder, device=device)
            "该数据库使用的模型（该模型必须是或继承自SentenceTransformer）。"
        elif isinstance(model, SentenceTransformer):
            self.model = model
        else:
            raise TypeError(f'期望：{SentenceTransformer}, 实际：{type(model)}')

        self.ebmapping: _docNodes = [
            node for tree in forest for node in tree.DFT()]
        "Eb节点映射，包含了数据库中所有的节点，可通过索引的方式直接顺序访问这些节点（不推荐）。"

        self.embeddings: Embedding = self.model.encode([
            node.data['content']
            for node in self.ebmapping
        ], normalize_embeddings=True)
        self.embeddings = self.embeddings.astype(np.double)
        "文本嵌入查询矩阵，用于计算查询相似度，不推荐读取。"

        self.c_methods: c_methods = ctypes.CDLL(const.c_source.vectordb)

    def insert(self, index: int, node: 'docNode'):
        self.ebmapping.insert(index, node)
        self.embeddings = np.insert(
            self.embeddings, index,
            self.model.encode(node.data['content']),
            axis=0
        )

    def query_by_faiss(self, text: str, count: int = 1, threshold: float = 0.5) -> Union['docNode', list['docNode']]:
        """
        从数据库中查询节点。（基于faiss的查询）
        - text: str, 查询的文本，作为相似性判断的根据。
        - count: int, 查询节点的数量。
        - threshold: float = 0.5, 查询阈值，低于阈值将被抛弃。
        """
        query_text = self.__find_chinese(text)
        if query_text == '':
            return None
        final_text = self.model.encode(
            query_text, normalize_embeddings=True)
        text_vector = np.array([final_text])
        # pytorch -> fairs
        dimension = self.embeddings.shape[-1]
        indexs = faiss.IndexFlatIP(dimension)
        indexs.add(self.embeddings)
        distance, indices = indexs.search(text_vector, count)
        indexs = list(indices.flatten())

        nodes = []
        for index in indexs:
            nodes.append(self.ebmapping[index])
        if len(nodes) == 1:
            return nodes[0]
        return nodes

    def query_by_cpp(self, text: str, count: int = 1, threshold: float = 0.5) -> Union['docNode', list['docNode']]:
        query_text = self.__find_chinese(text)
        if query_text == '':
            return None
        final_text = self.model.encode(
            query_text, normalize_embeddings=True)
        text_vector = np.array(final_text, dtype=np.double)
        query_matrix = self.embeddings.ctypes.data_as(np.ctypeslib.ndpointer(
            ctypes.c_double, self.embeddings.ndim, self.embeddings.shape))
        query_text = text_vector.ctypes.data_as(np.ctypeslib.ndpointer(
            ctypes.c_double, text_vector.ndim, text_vector.shape))
        index = self.c_methods.search(
            query_matrix, query_text, *self.embeddings.shape)

        nodes = []
        nodes.append(self.ebmapping[index])
        if len(nodes) == 1:
            return nodes[0]
        return nodes

    @debugger.method_time_comparison(query_by_cpp, 10)
    def query(self, text: str, count: int = 1, threshold: float = 0.5) -> Union['docNode', list['docNode']]:
        """
        从数据库中查询节点。
        - text: str, 查询的文本，作为相似性判断的根据。
        - count: int, 查询节点的数量。
        - threshold: float = 0.5, 查询阈值，低于阈值将被抛弃。
        """
        query_text = self.__find_chinese(text)
        if query_text == '':
            return None
        final_text = self.model.encode(
            query_text, normalize_embeddings=True)
        text_vector = np.array([final_text])
        # pytorch -> faiss
        similarity = (self.embeddings @ text_vector.T).flatten()
        indexs = torch.from_numpy(similarity).topk(count).indices
        indexs = list(indexs)
        _max = similarity.max()
        if _max < threshold:
            return None
        for position, index in enumerate(indexs):
            if similarity[index] < threshold:
                del indexs[position]

        nodes = []
        for index in indexs:
            nodes.append(self.ebmapping[index])
        if len(nodes) == 1:
            return nodes[0]
        return nodes

    def __query(self, text: str) -> int:
        text = self.model.encode(text, normalize_embeddings=True)
        text_vector = np.array([text])
        index = (self.embeddings @ text_vector.T).argmax()
        return index

    def __find_chinese(self, text: str):
        pattern = re.compile(r'[^\u4e00-\u9fa50-9]+')
        chinese = re.sub(pattern, '', text)
        return chinese

    def delete(self, text_or_index: str | int) -> Self:
        """
        **警告**：不推荐使用该方法删除节点，其具有不可预见性。
        删除节点，并返回被删除的节点。
        - text_or_index: str | int, 为文本时，根据相似度删除节点，为索引时删除对应索引位置的节点。
        """

        if isinstance(text_or_index, int):
            index = text_or_index
        elif isinstance(text_or_index, str):
            index = self.__query(text_or_index)
        else:
            raise ValueError('只能接收索引或字符串。')

        self.embeddings = np.delete(self.embeddings, index, axis=0)
        result = self.ebmapping[index]
        del self.ebmapping[index]

        return result

    def export(self, name: str, directory: str = './'):
        return super().export(name, directory)

    def save(self):
        return super().save()

    def load(path: str) -> Self:
        return super().load(path)

    @property
    def data(self):
        return super().data



import os
import re
import abc
import pickle
from typing import Any, Self, TypeAlias


def __import():
    global Node, Tree
    from nodoc import Node
    from nodoc import Tree

_Forest: TypeAlias = list['Tree']

class dataBase(metaclass=abc.ABCMeta):

    def __init__(self) -> None:
        self._data: Any
        self.__path: str = ""

    @abc.abstractmethod
    def export(self, name: str, directory: str = './'):
        """
        将数据库导出到对应的路径。
        - name: str, 数据库的导出名称（只能是数字、字母和下划线的组合）。
        - directory: str = './', 数据库的导出目录。
        """
        export_pattern = re.compile(r'([0-9]|[a-z]|[A-Z]|_)+', re.UNICODE)
        if not export_pattern.match(name):
            raise ValueError('Database names: combination of "0~9", "a~z", "A~Z" and "_."')
        
        directory = os.path.abspath(directory) + '\\' # 预处理目录
        with open(directory + name + '.nodocdb', 'wb+') as file:
            pickle.dump(self, file)

    @staticmethod
    @abc.abstractmethod
    def load(path: str) -> 'dataBase':
        """
        将数据库从磁盘载入内存。
        - path: str, 数据库的导入路径。
        """
        with open(path, 'rb+') as file:
            return pickle.load(file)

    @abc.abstractmethod
    def save(self) -> None:
        """
        保存数据库（注：必须是由load方法载入的数据库才能使用save方法）
        """
        if self.__path == "":
            raise AttributeError("数据库从未存储至磁盘。")
        path = os.path.abspath(self.__path)
        path = os.path.splitdrive(path)
        directory = path[0]
        file = path[1]
        name = os.path.splitext(file)[0]
        self.export(name, directory)

    @property
    @abc.abstractmethod
    def data(self) -> Any:
        """数据库的信息"""
        return self._data


from collections import deque
from typing import Any, Callable, Iterable, Literal, Protocol, Self, Sequence, TypeAlias, Union
import abc


class node_return_bool(Protocol):
    
    def __call__(self, node: Any) -> bool:
        ...

def auto_tree_updater(self: 'Node', node: 'Node'):
    def updater(node: 'Node', tree: 'Tree'):
        node._tree = tree
        if hasattr(node, 'seeker'):
            updater(node.seeker, tree)
        return
    
    if self._tree is None and node._tree is not None:
        self._tree = node._tree
        if hasattr(self, 'seeker'):
            updater(self.seeker, node._tree)
    
    elif self._tree is not None and node._tree is None:
        node._tree = self._tree
        if hasattr(self, 'seeker'):
            updater(node.seeker, self._tree)

    elif self._tree is None and node._tree is None:
        node.seeker = self
    return

class Node(metaclass=abc.ABCMeta):

    def __init__(self, **data) -> None:
        """
        节点
        - 关键字参数:
          - **data: Any, 节点保存的数据
        - 属性:
          - parent: Node, 该节点的父节点
          - left: Node, 该节点的左节点
          - right: Node, 该节点的左节点
          - children: list[Node], 该节点的所有子节点
          - visited: bool, 该节点是否被访问
        """
        self._parent: Node | None = None          # 父节点
        self._left: Node | None = None            # 左节点
        self._right: Node | None = None           # 右节点
        self._children: Nodes = Nodes([])       # 子节点列表
        self._data = data                  # 节点数据
        self._visited: bool = False        # 是否被访问
        self._count: int = 1               # 包括自己在内的节点数量（包含深度）
        self._depth: int = 0               # 节点的深度
        self._tree: Tree | None = None
        self.seeker: Node

    @property
    def tree(self) -> Union['Tree', None]:
        return self._tree
    
    @tree.setter
    def tree(self, tree: 'Tree'):
        if not isinstance(tree, Tree):
            raise TypeError(f'期望：{Tree}，实际：{type(tree)}')
        self._tree = tree

    @property
    def depth(self) -> int:
        return self._depth

    @depth.setter
    def depth(self, value: 'int'):
        if not isinstance(value, int):
            raise TypeError(f'期望：{int}，实际：{type(value)}')
        self._depth = value

    @property
    def visited(self) -> bool:
        "是否被访问"
        return self._visited

    @visited.setter
    def visited(self, value: bool):
        self._visited = value

    @property
    def parent(self):
        """
        父节点
        - setter:
          - node: `Node`，设置该节点的父亲节点
        """
        return self._parent

    @parent.setter
    def parent(self, node: 'Node'):
        if not isinstance(node, Node):
            raise TypeError(f'期望：{Node}，实际：{type(node)}')

        if self._parent:
            self._parent._children.remove(self)

        self._parent = node
        self._parent._children.append(self)
        self.depth = self._parent.depth + 1  # 节点深度 + 1
        auto_tree_updater(self, node)

    @property
    def left(self) -> Union['Node', None]:
        """
        兄弟节点-左
        - node: `Node`，设置该节点的左节点
        """
        return self._left

    @left.setter
    def left(self, node: 'Node'):
        if not isinstance(node, Node):
            raise TypeError(f'期望：{Node}，实际：{type(node)}')

        self._left = node
        if self._left.right is not self:  # 同时防止递归过深
            self._left.right = self
            auto_tree_updater(self, node)

    @property
    def right(self) -> Union['Node', None]:
        """
        兄弟节点-右
        - node: `Node`，设置该节点的右节点
        """
        return self._right

    @right.setter
    def right(self, node: 'Node'):
        if not isinstance(node, Node):
            raise TypeError(f'期望：{Node}，实际：{type(node)}')

        self._right = node
        if self._right.left is not self:  # 同时防止递归过深
            self._right.left = self
            auto_tree_updater(self, node)
            

    @property
    def children(self) -> 'Nodes':
        "子节点"
        return self._children

    @children.setter
    def children(self, value):
        self._children = value

    @property
    def data(self):
        "节点数据"
        return self._data

    @data.setter
    def data(self, value: Any):
        self._data = value

    @property
    def count(self) -> int:
        "节点数量"
        return self._count

    @count.setter
    def count(self, value: int):
        self._count = value

    @property
    def type(self):
        "节点类型"
        return type(self._data)

    @abc.abstractmethod
    def get_route(self, endpoint: 'Node', condition: node_return_bool = lambda node: True) -> 'Nodes':
        """
        获取子父节点的路由 -> 从该节点到目标节点之间的所有节点（不包括该节点）。
        - endpoint: Node，该节点的目标节点。
        - condition: Callable, 每个节点
        """
        if not isinstance(endpoint, Node):
            raise TypeError(f'期望：{Node}，实际：{type(endpoint)}')
        elif self is endpoint:
            return None
        result: list['Node'] = []

        if self.depth < endpoint.depth:
            current_node = endpoint
            endpoint = self
            reverse = False
        else:
            current_node = self
            reverse = True

        while (
            current_node.parent is not None and
            current_node is not endpoint
        ):

            current_node = current_node.parent
            if condition(current_node):  # 满足条件则追加该节点
                result.append(current_node)
        if reverse:
            result.reverse()
        if len(result) == 0:
            return None
        return Nodes(result)

    def __lshift__(self, other: Union['Node', None]):
        if isinstance(other, Node):
            return other.get_route(self)

    def __rshift__(self, other: Union['Node', None]):
        if isinstance(other, Node):
            return self.get_route(other)

    def __matmul__(self, other):
        if isinstance(other, Node):
            self.parent = other
            return other
        
        elif isinstance(other, Tree):
            self._tree = other
            return other

    def __and__(self, other):
        if isinstance(other, Node):
            self.right = other
            return Nodes([self, other])

        elif isinstance(other, Nodes):
            _ = self & other[0]
            other.prepend(self)


class Nodes(metaclass=abc.ABCMeta):

    def __init__(self, nodes: Union['Nodes', list[Node]]) -> None:
        self.data: deque[Node] = deque(nodes)

    def __matmul__(self, other):
        for node in self.data:
            _ = node @ other
        return other

    def __and__(self, other):
        if isinstance(other, Node):
            _ = self[-1] & other
            self.append(other)
        return other

    def __getitem__(self, index) -> Node:
        return self.data[index]

    def __iter__(self):

        for node in self.data:
            yield node  

    def prepend(self, node: Node):
        self.data.appendleft(node)

    def append(self, node: Node):
        self.data.append(node)

    def remove(self, node: Node):
        self.data.remove(node)

    def __str__(self) -> str:
        return "".join([str(node) for node in self.data])
    
    def __len__(self) -> int:
        return len(self.data)


class Tree(metaclass=abc.ABCMeta):

    def __init__(self, root: Node, name: str = '无名树') -> None:
        """
        实例化一个基本树对象。
        - root: Node, 树的根节点。
        - name: str, 树的名字，用于查询。
        """
        self.name: str = name
        self.root: Node = root
        self.current_node: Node = root
        self.__nodecount: int = 1
        
        for node in self.DFT():
            _ = node @ self

    @property
    def nodecount(self) -> int:
        return self.__nodecount

    """
    
    栈
    queue = [time1 -> a, time2 -> b...]
    以下循环：直到栈空
    栈出
    queue.popleft() 现在 queue 相当于 [time2 -> b, time3 -> c...]
    栈入
    queue.extend(nodeList) 现在 queue 相当于 [time2 -> b, time3 -> c...timeN1 -> Na, timeN2 -> Nb...]

    """

    def BFS(self, callback: Callable[[Node], bool] = lambda _: True) -> Node | None:
        """
        广度优先搜索（层次搜索）。
        - callback: Callable, 用于判断node是否符合指定条件。
        """
        visited = set()
        queue = deque([self.root])
        while queue:
            current_node = queue.popleft()

            if callback(current_node):
                return current_node  # 第一个符合条件的节点

            visited.add(current_node)

            queue.extend(child
                         for child in current_node.children
                         if child not in visited
                         and
                         child not in queue)
        return None  # 没有任何满足条件的节点

    @abc.abstractmethod
    def BFT(self, callback: Callable[[Node], bool] = lambda _: True) -> Nodes | None:
        """
        广度优先遍历（层次遍历）
        - callback: Callable, 用于判断node是否符合指定条件。
        """
        visited = set()
        queue = deque([self.root])
        result = []
        while queue:
            current_node = queue.popleft()

            if callback(current_node):
                result.append(current_node)  # 追加一个符合条件的节点

            visited.add(current_node)

            queue.extend(child
                         for child in current_node.children
                         if child not in visited
                         and
                         child not in queue)

        return Nodes(result)

    """
    
    递归
    从当前节点开始优先找子节点，子节点也优先找子节点，直到到底了都还没找到则返回父节点并找该节点的另一个子节点，以此类推

    """

    def DFS(self, node=None, callback: Callable[[Node], bool] = lambda _: True) -> Node | None:
        """
        深度优先搜索。
        - node: Node, 起点节点，默认为根节点。
        - callback: Callable, 用于判断node是否符合指定条件。
        """
        result = None
        if node is None:
            node = self.root

        if callback(node):
            return node  # 找到符合条件的节点

        node.visited = True
        for child in node.children:
            if not child.visited:
                result = self.DFS(child, callback)
            if result is not None:
                return result  # 在子树中找到符合条件的节点
        node.visited = False

        return None  # 在当前子树未找到符合条件的节点

    @abc.abstractmethod
    def DFT(self, node=None, callback: Callable[[Node], bool] = lambda _: True) -> Nodes:
        """
        深度优先遍历。
        - node: Node, 起点节点，默认为根节点。
        - callback: Callable, 用于判断node是否符合指定条件。
        """
        result = None
        if node is None:
            node = self.root

        result = []

        if callback(node):
            result.append(node)  # 找到符合条件的节点

        node.visited = True
        for child in node.children:
            if not child.visited:
                current_node = self.DFT(child, callback)
                result.extend(current_node)
        node.visited = False

        return Nodes(result)



from typing import Callable
import time
import sys
from nodoc import const


def method_time_comparison(compare_func: Callable, iter_time: int = 1000):
    """
    方法耗时比较。
    - compare_func: Callable, 被比较的方法（函数）。
    - iter_time: int, 迭代次数，默认1000次，次数越高，耗时越久同时精度越高。
    """
    def comparison(func: Callable):
        def wrapper(*args, **kwargs):

            duration_A = [] # 待比较函数的平均耗时列表
            duration_B = [] # 被比较函数的平均耗时列表
            for _ in range(iter_time):
                start_time = time.time()
                func(*args, **kwargs)
                duration = time.time() - start_time
                duration_A.append(duration)

                start_time = time.time()
                compare_func(*args, **kwargs)
                duration = time.time() - start_time
                duration_B.append(duration)
                
            result_A = func(*args, **kwargs)
            result_B = compare_func(*args, **kwargs)

            duration_A = sum(duration_A) / len(duration_A)
            duration_B = sum(duration_B) / len(duration_B)
            fast_or_slow = (duration_A / duration_B) > 1.0
            fast_or_slow_scale = [
                duration_B / duration_A,
                duration_A / duration_B
            ]
            func_A = func.__name__
            func_B = compare_func.__name__
            if hasattr(args[0], func_A):
                args = list(args)
                args[0] = args[0].__class__.__name__ + ' 实例'
            args = "\n- " + "\n- ".join([str(arg) for arg in args])
            func_name_width = max(len(func_A), len(func_B))
            outer_A_iter = list(str(duration_A).split('.')[1])
            outer_B_iter = list(str(duration_B).split('.')[1])
            outer_A = 1
            outer_B = 1
            for number_A, number_B in zip(outer_A_iter, outer_B_iter):
                if number_A == '0':
                    outer_A += 1
                if number_B == '0':
                    outer_B += 1
            outer_A += 3
            outer_B += 3
            outer = max(outer_A, outer_B)
            print(f"""
方法耗时比较，总测试次数：{iter_time}。
位置参数: {args}
关键字参数: {kwargs}
[函数：{func_A: <{func_name_width}}] 平均耗时：{duration_A:.{outer}f}s。
[函数：{func_B: <{func_name_width}}] 平均耗时：{duration_B:.{outer}f}s。
返回值{['不同', '相同'][result_A == result_B]}。
{func_A} {['快', '慢'][fast_or_slow]} {func_B} {fast_or_slow_scale[fast_or_slow]:.{outer}f} 倍。
""")

            return result_A, result_B

        return wrapper

    return comparison



from .base import Document
from .markdown import Markdown
from .pdf import PDF
from ._base import Message
from ._base import Checker

__all__ = [
    'Document',
    'Markdown',
    'PDF',
    'Message',
    'Checker'
]
try:
    from . import flow_processing as flowProcessing
    __all__.append('flowProcessing')
except:
    None



from nodoc.structure import docNode
from nodoc.structure import docTree
from nodoc.structure import Node
from nodoc.structure import Tree
from nodoc.database import vectorDB
from nodoc.document import Document
from nodoc.document import Markdown
from nodoc.document import PDF

import nodoc.debugger as debugger

__all__ = [
    'docNode',
    'docTree',
    'Node',
    'Tree',
    'vectorDB',
    'Document',
    'Markdown',
    'PDF',
    'debugger'
]
try:
    from nodoc.document import flowProcessing
    __all__.append('flowProcessing')
except:
    None


from typing import TypedDict, TypeAlias, Literal

ocrType: TypeAlias = Literal['text', 'title', 'header', 'reference', 'footer']

textRegion: TypeAlias = list[
    list[float, float],
    list[float, float],
    list[float, float],
    list[float, float],
]

bbox: TypeAlias = list[int, int, int, int]

class ocrResultElement(TypedDict):
    text: str
    confidence: float
    text_region: textRegion

ocrResult: TypeAlias = list[ocrResultElement]

class ocrDataElement(TypedDict):
    type: ocrType
    bbox: bbox
    res: ocrResult

ocrData: TypeAlias = list[ocrDataElement]


import abc
import time
from typing import Literal, Self, Union

from ._base import document_type
from ._base import Message
from ._base import Data
from ._base import Checker
from ._base import MessageInit
from ._base import MetaData
generalProperty = MessageInit.generalProperty
fontStyleProperty = MessageInit.fontStyleProperty

def __import():
    global docTree
    from nodoc import docTree




class Document(metaclass=abc.ABCMeta):

    def __init__(self, data: str = "", metadata: MetaData = None) -> None:
        if metadata is None:
            metadata = MetaData({
            'access': time.strftime("%Y-%m-%d %H-%M-%S"),
            'change': time.strftime("%Y-%m-%d %H-%M-%S"),
            'create': time.strftime("%Y-%m-%d %H-%M-%S"),
            'modify': time.strftime("%Y-%m-%d %H-%M-%S"),
            'custom': {},
            'filename': 'File',
        })
        self.__message: Message = self.normalize()
        self.__data: Data = Data(data)
        self.__metadata: MetaData = metadata

    @property
    def data(self) -> Data:
        return self.__data
    
    @property
    def metadata(self) -> MetaData:
        return self.__metadata

    @property
    def content(self) -> str:
        return self.__data.content

    @property
    def message(self) -> Message:
        return self.__message

    @message.setter
    def message(self, message: Message) -> None:
        self.__message = message

    @abc.abstractmethod
    def __document__(self):
        ...

    @abc.abstractmethod
    def normalize(self) -> Message:
        "文档标准化为消息。"

    @abc.abstractmethod
    def export(self, name: str, directory: str = './'):
        "文档的导出方法。"

    @staticmethod
    @abc.abstractmethod
    def load(path: str) -> 'Document':
        "文档的导入方法。"

    @staticmethod
    @abc.abstractmethod
    def load_from_message(message: Message) -> 'Document':
        """
        从消息中加载文档。
        - message: Message, 传入的消息。
        """

    @staticmethod
    def transform(source: 'Document', to: document_type) -> Union['Document', None]:

        match to:
            case 'markdown':
                from . import Markdown
                return Markdown.transform(source)
            # case 'html':
            #     from . import Html
            #     return Html.transform(source)
            # case 'pdf':
            #     from . import PDF
            #     return PDF.transform(source)
            # case 'word':
            #     from . import Word
            #     return Word.transform(source)
            # case 'ppt':
            #     from . import PPT
            #     return PPT.transform(source)
            # case 'excel':
            #     from . import Excel
            #     return Excel.transform(source)
        
    def treeify(self) -> 'docTree':
        """
        将文档树化。
        """

    def __lshift__(self, other):
        return self.__data << other

    def __str__(self) -> str:
        return self.__data.content



import os
import re
from time import strftime
from typing import Self, Union, Unpack
from uuid import uuid4

def __import():
    from nodoc import docTree


from .base import Data, Document, Message, Checker
from .base import generalProperty, fontStyleProperty
from .base import MetaData, time
# Use a pipeline as a high-level helper
INDENT = '\t'
END = '\n'
SPACE = '\x20'
TITLE_SIGN = '#'
REPLACE_FORMATTER = (
    (r'(\\)',  r'\\\1'),     # 转义\号
    (r'(\*)',  r'\\\1'),     # 转义*号
    (r'(\_)',  r'\\\1'),     # 转义_号
    (r'(\`)',  r'\\\1'),     # 转义`号
    (r'(\|)',  r'\\\1'),     # 转义|号
    (r'(!)',  r'\\\1'),      # 转义!号
    (r'(\[|\])',  r'\\\1'),  # 转义[]
    (r'(\(|\))',  r'\\\1'),  # 转义()
    (r'^(\x20*[0-9]+)(\.)(\x20)',  r'\1\\\2\3'),   # 有序列表
    (r'^(\x20*)(-|\*|\+)(\x20)',  r'\1\\\2\3'),    # 无序列表
    (r'^(\x20*)(#{1,6})(\x20)',  r'\1\\\2\3'),     # 标题
    (r'^(\x20*)(>)(.*)',  r'\1\\\2\3'),            # 引用
    (r'^(\x20*)(\*{3,}|-{3,}|_{3,})(\x20*)$',  r'\1\\\2\3'),    # 分隔线
    (r'^\s*$',  ''),
)

RECOGNIZE_FORMATTER = ()

def __import():
    global docTree
    from nodoc import docTree

class Markdown(Document):
    
    def __init__(self, data: str = "", metadata: MetaData = None) -> None:
        super().__init__(data, metadata)

    @staticmethod
    def escape(text: Union['Markdown', str]):
        """
        转义文本
        - text: str, 待转义文本
        """
        if isinstance(text, Markdown):
            text = text.content
        for pattern, repl in REPLACE_FORMATTER:
            text = re.sub(pattern, repl, text, flags=re.M)
        return text

    def next_line(self):
        "切换到下一行"
        _ = self.data << END

    def add_title(self, text: str, level: int):
        _ = self.data << TITLE_SIGN * level << SPACE << text << END * 2

    def add_text(self, text: str):
        text = self.escape(text)
        _ = self.data << text << END

    def add_comment(self, text):
        _ = self.data << f'<!-- {text} -->'
        if text == 'title':
            _ = self.data << END
        return text
    
    def add_fontStyle(self, style: fontStyleProperty, hash: str | None = None):
        bold_number = Checker("bold", "initial") >> style
        if not isinstance(bold_number, str):
            bold_number *= 100
        strList = str(""
                + f'text-align: {Checker("align", "initial") >> style};'
                + f'font-weight: {bold_number};'
                + f'color: {Checker("color", "initial") >> style};'
                + f'word-spacing: {Checker("font_distance", "initial") >> style};'
                + f'font-family: {Checker("font_family", "initial") >> style};'
                + f'font-size: {Checker("font_size", "initial") >> style};'
                + f'height: {Checker("height", "initial") >> style};'
                + f'width: {Checker("width", "initial") >> style};'
                + f'font-style: {"italic" if Checker("italic") >> style else "initial"};'
                + f'line_distance: {Checker("line-height", "5px") >> style};'
                + f'text-decoration: {"underline" if Checker("underline") >> style else "initial"};'
            ).split(';')
        if hash is None:
            for i in range(len(strList)):
                if not (strList[i].endswith('px') or strList[i].endswith('initial')):
                    if  not (strList[i].startswith('text-align')
                        or strList[i].startswith('font-family')
                        or strList[i].startswith('font-style')
                        or strList[i].startswith('text-decoration')
                        or strList[i] == ''):
                        strList[i] += 'px'
            return 'style="' + ";".join([x for x in strList if not x.endswith('initial')]) + ''
        _ = (self.data 
            << '<style>' << END
            << f".{hash} " 
            << '{' << END << INDENT
            << ";\n\t".join([x for x in strList if not x.endswith('initial')])[:-1]
            << '}' << END
            << '</style>' << END
            )

    def add_footer(self, text):
        _ = self.data << f'<!-- {text} -->' << END
        return text

    def normalize(self) -> Message:
        ...

    def export(self, name: str, directory: str = './'):
        directory = os.path.abspath(directory)
        with open(directory + '/' + name + '.md', 'w+', encoding='utf-8') as file:
            file.write(self.__str__())

    def treeify(self) -> 'docTree':
        from nodoc import docNode
        from nodoc import docTree
        node = docNode(kind = 'title', content = '')
        node.level = -1
        uuid = uuid4()
        tree = docTree(node)
        node @ tree
        message = self.message
        pipeline = message['pipeline']
        for order, unit in enumerate(pipeline):
            last_node: docNode = node
            node.order = order
            node: docNode = docNode(kind = unit('type'), content = unit['content'])
            match unit['type']:
                case 'text':
                    if last_node.isText:
                        self.message['pipeline'][node.order]['parent'] = uuid
                        (node & last_node) @ last_node.parent
                    elif last_node.isTitle:
                        node @ last_node

                case 'title':
                    node.level = unit['property']['level']
                    def bind_parent(nodeA: docNode, nodeB: docNode):
                        if nodeA.level == nodeB.level - 1:
                            self.message['pipeline'][nodeA.order]['parent'] = uuid
                            nodeA @ nodeB
                        elif nodeA.level == nodeB.level:
                            self.message['pipeline'][nodeA.order]['parent'] = uuid
                            nodeA @ nodeB.parent
                        elif nodeB.level == -1:
                            self.message['pipeline'][nodeA.order]['parent'] = uuid
                            nodeA @ nodeB
                        else:
                            bind_parent(nodeA, nodeB.parent)
                    if last_node.isText:
                        self.message['pipeline'][order]['parent'] = uuid
                        (node & last_node) @ last_node.parent
                    elif last_node.isTitle:
                        bind_parent(node, last_node)
            uuid = uuid4()
            self.message['pipeline'][order]['uuid'] = uuid
        return tree


    @staticmethod
    def transform(source: Document) -> 'Markdown':
        """
        静态方法，将任意受支持的文档类型转换为Markdown文档。
        - source: Document, 传入的文档。
        """
        document = Markdown.load_from_message(source.message)
        return document

    @staticmethod
    def load_from_message(message: Message) -> 'Markdown':
        """
        从消息中加载markdown。
        - message: Message, 传入的消息。
        """
        document: Markdown = Markdown(metadata=message['metadata'])
        document.message = message
        for unit in message['pipeline']:
            match unit['type']:
                case 'title':
                    if 'inherited' in unit['property']:
                        if isinstance(unit['property']['inherited'], str):
                            document.data << f'<span class="{unit["property"]["inherited"]}" style="'
                        else:
                            document.data << f'<span {document.add_fontStyle(unit["property"]["inherited"])}>'
                    else:
                        document.data << f'<span style="'
                    if True:
                        document.data << '">' << END
                    document.data << END
                    document.add_comment('title')
                    document.add_title(
                            Checker('content') >> unit, 
                            Checker('level') >> unit['property']
                        )
                    document.data << '</span>' << END
                case 'text':
                    if 'inherited' in unit['property']:
                        if isinstance(unit['property']['inherited'], str):
                            document.data << f'<span class="{unit["property"]}" style="'
                        else:
                            document.data << f'<span {document.add_fontStyle(unit["property"]["inherited"])}'
                    else:
                        document.data << f'<span style="'
                    if Checker('indent', None) >> unit['property'] is None:
                        document.data << '">' << END
                    else:
                        document.data << f"text-indent: {Checker('indent', 'initial') >> unit['property']}px" << '">' << END
                    document.data << END
                    document.add_comment('text')
                    document.add_text(
                        Checker('content') >> unit
                    )
                    document.data << '</span>' << END
                case 'latex':
                    if 'inherited' in unit['property']:
                        if isinstance(unit['property']['inherited'], str):
                            document.data << f'<span class="{unit["property"]}">' << END
                        else:
                            document.data << f'<span {document.add_fontStyle(unit["property"]["inherited"])}>' << END
                    else:
                        document.data << f'<span>' << END
                    document.data << END
                    _ = document.data << (Checker('content') >> unit)
                    document.data << '</span>' << END
                case 'image':
                    ...
                case 'header':
                    ...
                case 'footer':
                    ...
                case 'fontStyle':
                    ...
                
        return document


    @staticmethod
    def load(path: str) -> 'Markdown':
        with open(path, 'r+', encoding='utf-8') as file:
            text = file.read()
        return Markdown(text)

    def __document__(self):
        ...


from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="nodoc",
    version="1.0.2",
    author="HangBack",
    author_email="3402977740@qq.com",
    description="Layer-by-layer manipulation of document data with ease and efficiency.",

    # 项目主页
    url="https://github.com/HangBack/nodoc",

    # 你要安装的包，通过 setuptools.find_packages 找到当前目录下有哪些包
    packages=find_packages(),

    classifiers=[
        # 发展时期,常见的如下
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # 开发的目标用户
        'Intended Audience :: Developers',

        # 属于什么类型
        'Topic :: Software Development :: Build Tools',

        # 许可证信息
        'License :: OSI Approved :: MIT License',

        # 目标 Python 版本
        'Programming Language :: Python :: 3.11'
    ],
    install_requires=[
        'click==8.1.7',
        'faiss-cpu==1.7.4',
        'fsspec==2023.12.2',
        'huggingface-hub==0.19.4',
        'joblib==1.3.2',
        'nltk==3.8.1',
        'numpy==1.26.2',
        'packaging==23.2',
        'pandas==2.1.4',
        'Pillow',
        'psutil==5.9.6',
        'python-dateutil==2.8.2',
        'pytz==2023.3.post1',
        'regex==2023.10.3',
        'safetensors==0.4.1',
        'scikit-learn==1.3.2',
        'scipy==1.11.4',
        'sentence-transformers==2.2.2',
        'sentencepiece==0.1.99',
        'six==1.16.0',
        'threadpoolctl==3.2.0',
        'tokenizers==0.15.0',
        'torch==2.1.1',
        'torchaudio==2.1.1',
        'torchvision==0.16.1',
        'tqdm==4.66.1',
        'transformers==4.36.1',
        'triton==2.1.0',
        'tzdata==2023.3'
    ],
    python_requires = ">=3.11, <3.12",
    long_description=long_description
)
